@@ -14,4 +14,22 @@
 # limitations under the License.
 #===============================================================================
 
-__all__ = ['svm', 'prims']
+import platform
+if "Windows" in platform.system():
+    import os
+    import sys
+    import site
+    path_to_env = site.getsitepackages()[0]
+    path_to_libs = os.path.join(path_to_env, "Library", "bin")
+    if sys.version_info.minor >= 8:
+        os.add_dll_directory(path_to_libs)
+    os.environ['PATH'] += os.pathsep + path_to_libs
+
+try:
+    import onedal._onedal_py_dpc as _backend
+    _is_dpc_backend = True
+except ImportError:
+    import onedal._onedal_py_host as _backend
+    _is_dpc_backend = False
+
+__all__ = ['primitives', 'svm']
